<?php
  /**
   *
   * CiviNode Utility Functions
   *
   * These include wrappers for common CRM related tasks like
   * determining group membership, going from NID to CID and
   * back, and doing loads.
   *
   */

  //lowlevel routines to bypass the normal API.
  //all of these routines should be treated as private
  //with extreme prejudice.
require_once 'lowlevel/civinode_groups.inc';

//Some magic numbers that Lobo won't code compatibly for us (rant rant)
define('CIVINODE_MAGIC_ACTION_NONE',                0);
define('CIVINODE_MAGIC_ACTION_ADD',                 1);
define('CIVINODE_MAGIC_ACTION_UPDATE',              2);
define('CIVINODE_MAGIC_ACTION_VIEW',                4);
define('CIVINODE_MAGIC_ACTION_DELETE',              8);
define('CIVINODE_MAGIC_ACTION_BROWSE',             16);
define('CIVINODE_MAGIC_ACTION_ENABLE',             32);
define('CIVINODE_MAGIC_ACTION_DISABLE',            64);
define('CIVINODE_MAGIC_ACTION_EXPORT',            128);
define('CIVINODE_MAGIC_ACTION_BASIC',             256);
define('CIVINODE_MAGIC_ACTION_ADVANCED',          512);
define('CIVINODE_MAGIC_ACTION_PREVIEW',          1024);
define('CIVINODE_MAGIC_ACTION_FOLLOWUP',         2048);
define('CIVINODE_MAGIC_ACTION_MAP',              4096);
define('CIVINODE_MAGIC_ACTION_PROFILE',          8192);
define('CIVINODE_MAGIC_ACTION_MAX_ACTION',      16383);



/**
 * Initialization checks
 *
 *
 */
function civinode_check_init(){
  if (function_exists('civicrm_initialize')) {
    civicrm_initialize(TRUE);
    return TRUE;
  }
  return FALSE;
}


/**
 * Check for Voter component
 */
function civinode_voter_install() {
	if (!civinode_check_init())
	  return FALSE;
	if (function_exists('crm_voter_create_campaign'))
	  return TRUE;
	else
	  return FALSE;
}

/**
 * Check for admin level access for this operation
 */
function civinode_check_admin_access($op = 'view'){
  global $user;
  //user_access is heavily cached, so we don't
  //worry about doing any more of it
  if ( user_access('administer CiviCRM', $user))
    return TRUE;
  if ( $op == 'view' and user_access('view all contacts'))
    return TRUE;
  if ( ($op == 'update' or $op == 'delete') and  user_access('edit all contacts'))
    return TRUE;
    
  return FALSE;

}


/**
 * Error handling. We will want to keep error info
 * around so we can pass it to UI, etc. The setter
 * will do this.
 *
 * @defgroup Error_Handlers
 */

/**
 * Error testing and logging routine.
 *
 * @ingroup Error_Handlers
 * @param mixed $obj value returned by CiviCRM API that may be error.
 * @return boolean TRUE if $obj is a CiviCRM error.
 */
function civinode_check_error($obj = NULL){
  //Reset the error
  civinode_clear_error();
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  if (!$obj) {
    $obj = _civinode_get_last_error();
  }
  //Docs are wrong... :-(
  //if(crm_is_error($obj)){
  if($obj and is_a($obj, 'CRM_Core_Error')){
    $msg = $obj->_errors[0]['message'];
    //may want to record the calling URL as well...
    $trace = debug_backtrace();
    $caller = $trace[1]['function'] . "@" . $trace[1]['line']; 
    watchdog('CRM', "$msg ($caller)", WATCHDOG_ERROR);
    _civinode_set_error($msg);
    return TRUE;
  }
  else
    return FALSE;

}

/**
 * Get the text of the last error checked via civinode_check_error
 * 
 * @ingroup Error_Handlers
 * @return string Error string.
 */
function civinode_get_last_error(){
  return _civinode_set_error();
}

/**
 * Clear the last error saved.
 *
 * @ingroup Error_Handlers
 * 
 */
function civinode_clear_error(){
  _civinode_set_error(NULL, TRUE);
}


function _civinode_set_error($value = NULL, $reset = FALSE){
  static $error;
  $last_error = $error;
  if($reset){
    $error = NULL;
    return NULL;
  }
  if(!$value)
    return $last_error;
  else{
    $error = $value;
    return $value;
  }

}


/**
 * Non destructively see if you can get the last error
 * off the error stack, if there is one.
 * 
 * @ingroup Error_Handlers
 * @return  for now, a PEAR error object
 */
function &_civinode_get_last_error(){
  $last_error = NULL;
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return "FATAL: CiviCRM not installed";
  }

  $error_stack =& CRM_Core_Error::singleton();
  //For now, let's see how "deep the rabbit hole goes" and
  //return just the tip of the stack.
  if ($error_stack->hasErrors()) {
    $errors = $error_stack->getErrors();
    $lastError = array_pop($errors);
  }
  
  return $last_error;
}


/**
 * Metadata about contacts
 * 
 * @defgroup CiviCRM_MetaData
 */


/**
 * Get a list of profile names.
 *
 * @ingroup CiviCRM_MetaData
 * @param array Search criteria for profiles (UNUSED).
 * @return array of int => string
 */
function civinode_get_profiles($params = NULL){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  //Note: as of 1.4, we ignore params
  //We may want to use the auth table to reduce this
  //list (i.e., you don't have the right, you can't
  //assign it.
  $profiles = crm_uf_get_profile_groups();
  //winnow it down here later...
  return $profiles;
}

/**
 * Get the human-readable name of a profile.
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @return string
 */
function civinode_get_profile_title($pid){
  static $cache;
  if(isset($cache[$pid]))
    return $cache[$pid];
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $cache[$pid] = crm_uf_get_profile_title($pid);
  return $cache[$pid];
}



/**
 * Get the right profile ID for displaying this drupal user.
 *
 * @ingroup CiviCRM_MetaData
 * @param int Drupal UID
 * @return int CiviCRM Profile ID
 */
function civinode_get_default_profile_id($uid = 0){
  $pid = variable_get('civinode_default_pid', 1);
  return $pid; //TO DO: make this settable for the individual
}


/**
 * Generate a Forms API structure for a Profile
 *
 */
function civinode_get_profile_form($pid, &$form, $params = array()) {
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  $meta = civinode_get_profile_metadata($pid);
  if (!$meta)
    return FALSE;
  if (!$form or !is_array($form))
    $form = array();
  //We may need to retrieve some non-standard fields...
  $extra_fields = array();
  if ($params['contact_id']) {
    foreach ($meta as $item) {
      $extra_key = $item['name'];
      if (!isset($params[$extra_key])) {
        $extra_fields[$extra_key] = 1;
      }
    }
    if ($extra_fields) {
      $extra_params =
        array('contact_id' => $params['contact_id']);
      $extra_info = crm_contact_search($extra_params, $extra_fields,
				       NULL, 0, 1);
      if (isset($extra_info[0]) and is_array($extra_info[0])) {
	$data = array_shift($extra_info[0]);
	foreach ($extra_fields as $fld_name => $dummy) {
	  if (isset($data[$fld_name])) {
            $val = $data[$fld_name];
            if (is_string($val) and $val[0] == chr(1)) {
              //we have an encoded checkbox, explode it
              $boxes = explode(chr(1), $val);
              $val = array();
              foreach ($boxes as $box) {
                $val[$box] = $box;
              }
            }
	    $params[$fld_name] = $val;
          }
	}
      }
      
    }
  }
  //Now, we build the form
  foreach ($meta as $item) {
    $html_type = $item['html_type'];
    //Valid types listed in CRM_Custom_Form_Field
    //support will be incremental, since I don't
    //know all the formats yet
    //  array(  'Text' => 'Text', 'Select' => 'Select', 
    //          'Radio' => 'Radio', 'CheckBox' => 'CheckBox',
    //          'Multi-Select' => 'Multi-Select'),
    //  also: TextArea, Date, some geographical
    $name = $item['name'];
    $label = $item['title'];
    
    switch ($html_type) {
      case 'CheckBox': //actually, a set of them
        //remove the '---' item
        unset($item['options'][0]);
        $form[$name] =
          array('#title' => $label,
                '#type' => 'checkboxes',
                '#options' => $item['options'],
                );
        break;
      case 'Radio': //civicrm 1.6 is a little odd
        //remove the '---' item
        if ($item['options']) {
          unset($item['options'][0]);
          $form[$name] =
            array('#title' => $label,
                  '#type' => 'radios',
                  '#options' => $item['options'],
                  );
        }
        else {
          //this is just a free-standing boolean.
          //We render it as a single checkbox
          $form[$name] =
            array('#title' => $label,
                  '#type' => 'checkbox',
                  );          
        }
        break;
      case 'Select':
        $form[$name] =
          array('#title' => $label,
                '#type' => 'select',
                '#options' => $item['options'],
                );
        break;
      case 'Text':
      //we deliberately fall through to default, since textfield is the
      //default rendering
      default:
        $form[$name] =
          array('#title' => $label,
                '#type' => 'textfield',
                );
        if ($item['attributes']) {
          $size = $item['attributes']['size'];
          $max = $item['attributes']['maxlength'];
          if ($size) {
            $form[$name]['#size'] = $size;
            if ($max)
              $form[$name]['#maxlength'] = $max;
          }
          
        }
        break;
    }
    if (isset($params[$name])) {
      //Note: this may need to be moved into the switch
      //if some items require special processing.  Dates
      //may well need this.
      $form[$name]['#default_value'] = $params[$name];
    }
    
    if ($item['help_post'])
      $form[$name]['#description'] = $item['help_post']; 
  }

}

/**
 * Save back profile related data.
 *
 * This function should be called in your FAPI submit function
 *
 */
function civinode_save_profile_fields($pid, $uid, $params) {
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  //We need to flatten the array valued item that comes back
  //for a set of checkboxes
  $old_params = $params;
  foreach ($old_params as $key => $val) {
    if (is_array($val)) {
      //CiviCRM does a weird trick with Ctrl-A...
      //$params[$key] = chr(1) . implode(chr(1), array_values($val)) . chr(1);
      $new_val = array();
      foreach (array_values($val) as $key_val) {
        if ($key_val) {
          $new_val[$key_val] = 1;
        }
      }
      $params[$key] = $new_val;
    }
  }
  $fields =  crm_uf_get_profile_fields($pid);
  //We need to work around a bug that will cause the display_name to
  //unset.  This will be fixed in 1.8, so this code will no longer
  //be needed once we are no longer supporting 1.6 and 1.7
  $contact_params = array('contact_id' => $uid);
  $disp_fields = array('first_name' => 1, 'last_name' => 1, 'middle_name' => 1);

  $contact = crm_fetch_contact($contact_params, $disp_fields);
  if (!civinode_check_error($contact)) {
    foreach ($disp_fields as $fld => $dummy) {
      if (!isset($params[$fld]) and isset($contact[$fld]))
        $params[$fld] = $contact[$fld];
    }
  }
  //End of display_name fix
  crm_create_profile_contact( $params, $fields, $pid, $uid);
}

/**
 * Get the appropriate meta data for this PID and operation
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of field_name => array
 */
function civinode_get_profile_metadata($pid, $op = ''){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  static $cache, $exportable;
  $action = CRM_CORE_ACTION_VIEW;
  switch($op){
  default:
    //We will translate Drupal ops into actions here
    $action = CRM_CORE_ACTION_VIEW;
    break;
  }
  //structure of cache: first by pid, then ACTION
  if(isset($cache[$pid])){
    if(isset($cache[$pid][$action]))
      return $cache[$pid][$action];
  }
  else
    $cache[$pid] = array();
    //Gather info about field structure
    require_once 'CRM/Contact/BAO/Contact.php';
    require_once 'CRM/Utils/Type.php';
    $dummy = new CRM_Contact_BAO_Contact();
    $exportable = $dummy->exportableFields();

    //what should visibility be set to??
    $raw_data = crm_uf_get_profile_fields($pid, FALSE, $action);
    $keys = array_keys($raw_data);
    foreach ($keys as $key) {
      if (isset($exportable[$key])) {
        //is it a custom field?
        if ($exportable[$key]['custom_field_id']) {
          $params = array('id' =>  $exportable[$key]['custom_field_id']);
          $obj = crm_get_custom_field( $params );
          if (!civinode_check_error($obj)) {
            $values = crm_get_option_values($obj);
            $options = array();
            if ($values) {
              if (!$raw_data[$key]['is_required']) {
                $options = array('0' => t('---') );
              }
              foreach ($values as $val_key => $val) {
                $options[$val['value']] = $val['label']; 
              }
              $raw_data[$key]['options'] = $options;
            }
          }
        }
        //Note: type for built-in, data_type and html_type for custom
        $data_type = $exportable[$key]['data_type'] ?
          $exportable[$key]['data_type'] :
          CRM_Utils_Type::typeToString($exportable[$key]['type']);
        $raw_data[$key]['type'] = $data_type;
        $raw_data[$key]['html_type'] = $exportable[$key]['html_type'] ?
          $exportable[$key]['html_type'] : _civinode_html_typer($data_type);
      }
    }
    //$title = crm_uf_get_profile_title($pid);
    $cache[$pid][$action] = $raw_data;
  
    return $cache[$pid][$action];
}

function _civinode_html_typer($type) {
  switch ($type) {
    case 'Boolean':
      return 'Radio';
    default:
      return 'Text';
  }
}


/**
 * This function is a nasty hack to get around the way location-related
 * fields get handled.  It returns a map from what names are used for 
 * fields in a node or the assoc array calls in the CRM API,
 * into the names used internally by CRM for formatting info.
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of string 
 */
function civinode_get_profile_field_list($profile_id, $op = 'view'){
  static $cache;
  $cache_key = "$op::$profile_id";
  if(isset($cache[$cache_key]))
    return $cache[$cache_key];
  $meta_data = civinode_get_profile_metadata($profile_id, $op);
  $fields = array();
  if($meta_data){
    $raw_fields = array_keys($meta_data);
    foreach($raw_fields as $raw_field){
      list($base, $loc_type_id) = explode('-', $raw_field);
      //does this preserve order??
      $fields[$base] = $raw_field;
    }
  }
  $cache[$cache_key] = $fields;
  return $fields;
}

/**
 * Get the appropriate meta data for this field for this PID and operation
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of string => mixed
 */
function civinode_get_field_info($pid, $fld_name, $op = 'view'){
  //This function is a candidate for yet more
  //caching, but see if we are spending lots of time here
  //first...

  $metadata = civinode_get_profile_metadata($pid, $op);
  //This would be easy, except that CRM plays games with locations.
  //Since it does, we will need to deal with multiple locations
  //and convert to the "base form"
  
  //First, maybe we won't have to deal with this problem
  if(isset($metadata[$fld_name]))
    return $metadata[$fld_name];
  //OK, not so lucky.  See if we are dealing with a location
  //dependent item
  $keys = array_keys($metadata);
  $real_key = "";
  foreach($keys as $key){
    list($base, $loc_type_id) = explode('-', $key);
    if($base == $fld_name){
      $real_key = $base;
      break;
    }
  }
  if($real_key)
    return $metadata[$real_key];

  return NULL; //we don't recognize this key
}


/**
 * Get a Drupal user's CRM contact_id.  Caching wrapper for crm_uf_get_match_id
 * 
 * @ingroup CiviCRM_MetaData
 * @param int Drupal UID
 * @return int CiviCRM Profile ID
 *
 */
function civinode_util_get_user_cid($uid){
  if(!$uid)
    return 0;
  static $cid_cache;
  if(!isset($cid_cache[$uid])){
    if (!civinode_check_init()){
      watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
      return FALSE;
    }
    $cid = crm_uf_get_match_id($uid);
    //No user should lack a cid, so we
    //may be registering.  Don't cache
    //to cover that case.
    if ($cid)
      $cid_cache[$uid] = $cid;
    else
      return 0;
  }
  return $cid_cache[$uid];
}


/**
 * Loads contact data using a profile as a filter
 *
 * @ingroup CiviCRM_MetaData
 * @param int $cid CRM Contact ID
 * @param int $pid CRM Profile ID
 * @return array assoc. array of contact data
 */
function civinode_util_load_cdata($cid, $pid = 0){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $params = array('contact_id' => $cid);
  if ($pid) {
    $return_properties = array();
    $keys = civinode_get_profile_field_list($pid, 'view');
    foreach (array_keys($keys) as $key) {
      $return_properties[$key] = 1;
    }
  }
  else {
    $return_properties = NULL;
  }

  $data = crm_fetch_contact($params, $return_properties);
  return $data;
}


/**
 * Generate a nice link back to the contact
 *
 * @param unknown_type $contact_id
 * @param unknown_type $format
 * @param unknown_type $params
 * @return unknown
 */
function civinode_util_link_to_crm($contact_id, $format = NULL, $params = array()) {
  $data = civinode_util_load_cdata($contact_id);
  if (!$data)
    return NULL;
  $new_params = array();
  if ($params) {
    foreach ($params as $key => $text) {
      $val_fld = _civinode_strip_t_fld_type($key);
      $new_params[$key] = $data[$val_fld];
    }
  }
  if (!$format) {
    $format = '%display_name';
    $new_params['%display_name'] = $data['display_name'];
  }
  $text = t($format, $new_params);
  //http://design/civicrm/contact/view?reset=1&cid=1624
  //link allowing HTML
  $link = l($text, "civicrm/contact/view", array(), "reset=1&cid=$contact_id", NULL, FALSE, TRUE);
  return $link;
}

function _civinode_strip_t_fld_type($fld) {
  //remove that first char
  $first_char = substr($fld, 0, 1);
  if ($first_char == '%' or  
      $first_char == ':' or 
      $first_char == '!')
      return substr($fld, 1); //rest of string
   else 
      return $fld;
}


/**
 * Information about groups
 * @defgroup CiviCRM_Groups
 */


/**
 * List groups by various criteria
 *
 * @ingroup CiviCRM_Groups
 * @param array $params assoc array of search params.  NULL is all groups.
 * @return array of int => string
 */
function civinode_get_groups($params = NULL){
  //we just wrap crm_get_groups
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  $groups = crm_get_groups($params);
  if(civinode_check_error($groups))
    return NULL;
  $group_list = array();
  foreach($groups as $gdata){
    $group_list[$gdata->id] = $gdata->title;
  }

  return $group_list;
  
}


/**
 * List groups for a contact.
 *
 * @ingroup CiviCRM_Groups
 * @param int $cid CiviCRM Contact ID
 * @return array of int => string
 */
function civinode_get_groups_for_cid($cid) {
  static $group_cache;
  if (!is_array($group_cache))
    $group_cache = array();
  $groups = array();
  if (!isset($group_cache[$cid])) {
    if (!civinode_check_init()){
      watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
      return FALSE;
    }
    $list = CiviNode_LL_getContactGroup($cid, 'Added');
    //$list = crm_contact_get_groups($cid, 'Added');
    foreach ($list as $eid => $data) {
      $groups[] = $data['group_id'];
    }
    $group_cache[$cid] = $groups;
  }
  else
    $groups = $group_cache[$cid];

  return $groups;
}


/**
 * Wrapper to add a contact to a group.
 *
 * @ingroup CiviCRM_Groups
 * @param int $group_id CiviCRM Group ID
 * @param int $cid CiviCRM Contact ID
 * @param string $source source to save to CRM data store.
 * @return boolean TRUE if the contact was successfully added.
 */
function civinode_add_contact_to_group($group_id, $contact_id, $source = 'civinode'){
  if (civinode_util_cid_in_group($contact_id, $group_id))
    return TRUE; //nothing to do.
  $group = civinode_get_group_by_id($group_id);
  if (!$group or civinode_check_error($group))
    return FALSE;
  $contact = crm_get_contact(array('contact_id' => $contact_id));
  if (!$contact or civinode_check_error($contact))
    return FALSE;
  //Work-around for a 1.4 bug: id is set here, but not contact_id
  if (!isset($contact->contact_id))
    $contact->contact_id = $contact->id;
  $contacts = array($contact);
  //'API' could be whatever you want, including the name of your module
  crm_add_group_contacts($group, $contacts, 'Added', 'API');
  return TRUE;
}


/**
 * Remove a contact from a group
 *
 * @ingroup CiviCRM_Groups
 * @param int $group_id CiviCRM Group ID
 * @param int $cid CiviCRM Contact ID
 * @return boolean TRUE if the contact was successfully removed (or was not in this group).
 */
function civinode_remove_contact_from_group($gid, $cid){
  $group = civinode_get_group_by_id($id);
  if (!$group)
    return FALSE;
  if (!civinode_util_cid_in_group($cid, $gid))
    return TRUE; //nothing to do, but we are also OK
  $to_remove = array();
  $cobj = crm_get_contact(array('contact_id' => $cid));
  if (civinode_check_error($cobj))
    return FALSE;
  $to_remove[] =& $cobj;
  $rslt = crm_delete_group_contacts($group, $to_remove, 'API');
  if (!$rslt)
    return TRUE; //success
  else
    return !civinode_check_error($rslt); //so we log the error  
    
}


/**
 * Get a group by name, creating it if need be
 *
 * @ingroup CiviCRM_Groups
 * @param string $name name of the group to retrieve or create. 
 * @param string $source what module created group (UNUSED?)
 * @param boolean $create if true and no group exists, create a group.
 * @return int The CiviCRM Group ID of the group (false/0 if failed)
 */
function civinode_get_group_by_name($name, $source = 'civinode',  $create = TRUE){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $groups = crm_get_groups(array('title' => $name));
  if(civinode_check_error($groups))
    return NULL;
  $gid = 0; //pessimism
  if (count($groups)) {
    //find an exact match
    foreach ($groups as $group) {
      if ($name == $group->title) {
        $gid = $group->id;
        break;
      }
    }

  }
  else if ($create) {
    //Need a fresh minted group
    $params = array('title' => $name,
                    'source' => $source,
                    'is_active' => 1);
    $group = crm_create_group($params);
    if(civinode_check_error($group))
      return NULL;
    $gid = $group->id;
  }

  return $gid;

}


/**
 * Wrapper to fetch a CRM group object
 *
 * @ingroup CiviCRM_Groups
 * @param int $gid CiviCRM Group ID
 * @return object the CiviCRM group object (NULL on error)
 */
function civinode_get_group_by_id($gid){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $groups = crm_get_groups(array('id' => $gid));
  if(civinode_check_error($groups))
    return NULL;
  //Is this a behavior change from 1.3???
  return count($groups) ? $groups[0] : NULL;
}



/**
 * Get target groups as allowed for the current user
 *
 * @ingroup CiviCRM_Groups
 * @return array of int => string.
 */
function civinode_get_drupal_allowed_groups() {
	//TO DO: this can be greatly sped up.
  $gids = civinode_get_current_target_groups('view');
  $groups = array();
  if (count($gids)) {
    foreach ($gids as $gid) {
      $group = civinode_get_group_by_id($gid);
      $groups[$gid] = $group->title;
    }
  }
  return $groups;
}

/**
 * Groups for this user
 *
 * @param string $op
 * @return array
 */
function civinode_get_current_target_groups($op = 'view') {
  civinode_check_init();
  $group_list = civinode_get_groups();
  return array_keys($group_list);
}


/**
 * Determines if a CRM group contains a contact.
 *
 * @ingroup CiviCRM_Groups
 * @param int $cid CRM Contact ID
 * @param int $gid CRM Group ID
 * @return bool TRUE if it belongs
 */
function civinode_util_cid_in_group($cid, $gid){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }

  //TO DO: use a search instead.

  //What groups does this cid know about?
  $list = CiviNode_LL_getContactGroup($cid, 'Added');
  //$list = crm_contact_get_groups($cid, 'Added');
  //AFAIR keys are gid
  $groups = array_keys($list);
  return in_array($gid, $groups);
}




/**
 * Wrapper for getting back an array of contact ids.
 *
 * @ingroup CiviCRM_Groups
 * @param int $gid CiviCRM group ID
 * @param int $start index of first record 
 * @param int $num_recs number of records to pull. 0 is "all records"
 *
 */
function civinode_util_group_contacts($gid, $start = 0, $num_recs = 0){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $group = crm_get_groups(array('id' => $gid));
  $list = array();
  //establish a limit, due to memory bugs
  //in 1.4.
  $limit = $num_recs ? $num_recs : 1000; 
  if(civinode_check_error($group))
    return NULL;
  if (!count($group))
    return array();
  $contacts = crm_get_group_contacts($group[0],
                                     //array('id' => 1), //bare min please
                                     NULL, //default instead
                                     'Added', NULL, //status and sort
                                     $start, $limit); //up to first 10K records
  foreach($contacts as $contact){
    $list[] = $contact->contact_id;
  }
  return $list;

}


/**
 * Activity wrappers
 * 
 * @defgroup CiviCRM_Activity_History
 */


/**
 * Creates a simple activity history record
 *
 * @ingroup CiviCRM_Activity_History
 * @param int $cid  CRM Contact ID to associate with the activity
 * @param int $id_to_save  An integer key you want to save with the activity
 * @param string $activity_type  'Meeting' or other defined activity type
 * @param string $module_base  UI name of the module storing this stuff.
 * @param string $desc  Short description to display about the activity
 * @param int $group_id  An optional CRM group_id to associate with the activity
 * @param string $call_back  The name of a PHP function with the following
 *               prototype, returning a URL:
 *               
 *               function my_details_handler($it_to_save, $crm_activity_hist_id);
 * 
 *               You can use civinode_get_activity($crm_activity_hist_id) to get back
 *               an associative array with the data stored in the activity record.
 *                 
 *
 * @param int $date  Unix time stamp.  Default is now, i.e., time().
 * @return boolean  TRUE on success
 */
function civinode_create_activity($cid, $id_to_save, $activity_type,
                                  $module_base = 'civinode',
                                  $desc = "", $group_id = 0,
                                  $call_back = NULL, $date = NULL){
  $params = array('module' => $module_base,
                  'entity_table' => 'civicrm_contact',
                  'entity_id' => $cid,
                  'activity_summary' => $desc,
                  'activity_type' => $activity_type);
  if (!$date)
    $date = time();

  $date_str = date('Ymd000000', $date);
  $params['activity_date'] = $date_str;
  $params['activity_id'] = $id_to_save;
  //callback...
  //$url = url($link);
  if ($call_back)
    $params['callback'] = $call_back;
  $hist = crm_create_activity_history($params);
  return civinode_check_error($hist);
}


/**
 * Getter for activity history data
 *
 * @ingroup CiviCRM_Activity_History
 */
function civinode_get_activity($crm_activity_hist_id){
  $defaults = array();
  $params = array('id' => $crm_activity_hist_id);
  $activity = crm_get_activity_history_object($params, $defaults);
  if (civinode_check_error($activity))
    return NULL;
  return $defaults;
}



/**
 * Sample URL callback for activities
 *
 * @ingroup CiviCRM_Activity_History
 */
function  civinode_activity_callback($saved_id, $crm_activity_id) {
  //error_log("Call back was called");
  $defaults = civinode_get_activity($crm_activity_id);
  if (!civinode_check_error($defaults)) {
    //We are cheating: we assume that the module is really
    //a path
    if (isset($defaults['module'])) {
      $base = $defaults['module'] . "/activity/$crm_activity_id/$saved_id";
      return url($base);
    }
    
  }
}


//TO DO:  we may want to do synchronization via cron as well.  PITA. :-(

/**
 * Form API widgets and stuff
 *
 * @defgroup CiviNode_Form_Widgets
 */



/**
 * AJAX autocompleter handler for CiviCRM Contacts
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_contact_autocompleter($key, $val, $string){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), 
             WATCHDOG_ERROR);
    exit();
  }
  //first, require that we have at least 3 characters
  if (strlen($string) < 3)
    exit();

  //OR is really needed here. Lobo??
  $return_fields = array('display_name' => 1, 
                         'id' => 1,
                         'contact_type' => 1,
                         );
  
  $params['sort_name'] = "%$string%";
  //Check the key/val argument for special handling
  if ($key == 'gid' and is_numeric($val)) {
    $params['group'] = array($val => 1);

  }
  $rslt = crm_contact_search($params, $return_fields);
  $matches = array();
  foreach($rslt[0] as $id => $data){
    $id_str = " (" . $id . ")";
    $matches[$data['display_name'] . $id_str] = check_plain($data['display_name'] . $id_str);
  }
  //This assumes you are using
  //4.7 RC2 or later...
  print drupal_to_js($matches);
  exit();

}



/**
 * Returns a keyed array of id => 'useful display name' for all basic CRM types
 * It assumes that we want to check for access permissions for the global
 * Drupal user (for contacts and groups). This function is useful for presenting
 * selection options to users during input forms, etc.
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_get_allowed_options($type, $params = NULL) {
  static $cache;
  if(isset($cache[$type]) and !$params) {
    return $cache[$type];
  }
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  //See if we have voter available
  $voter_install = civinode_voter_install();
  
  $result = array();
  switch($type) {
  case 'relationship':
    $relationships = crm_get_relationship_types();
    while($relationship = array_pop($relationships)) {
      $result[$relationship->id] = $relationship->description;
    }
    break;
    
  case 'group':
    $result = civinode_get_drupal_allowed_groups();
    break;
      
  case 'profile':
    $result = crm_uf_get_profile_groups();
    break;
    
  case 'contact':
    $allowed_groups = civinode_get_drupal_allowed_groups();
    $groups = array();
    //we restrict to the requested groups if there are any.
    //0 is the "universal" group, so skip this if it is requested
    if ($params and isset($params['groups']) and !in_array(0, $params['groups'])) {
      foreach ($params['groups'] as $requested_group) {
        if (isset($allowed_groups[$requested_group]))
          $groups[$requested_group] = $allowed_groups[$requested_group];
      }
    }
    else
      $groups = $allowed_groups;
    foreach($groups as $gid => $title) {
       
      $tmp = crm_get_groups(array('id' => $gid));
      $group = array_pop($tmp);
      $group_contacts = crm_get_group_contacts($group, array('display_name'));
      foreach($group_contacts as $contact) {
        $result[$contact->contact_id] = $contact->display_name;
      }
    }
    break;
    
  case 'activity':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::activityType(true, 'id > 3');
    break;
    
  case 'tag':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::tag();
    break;
      
  case 'location':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::locationType(true);
    break;
    
  case 'campaign':
  	if ($voter_install) {
      $result = civinode_voter_campaign_selector();
  	}
  	break; 
  
  case 'canvass':
  	if ($voter_install) {
  		$result = civinode_voter_canvass_selector();
  	}
  	break; 
  
    
    /* disabled for now
     case 'field':
     $fields = crm_get_class_properties('Individual', 'custom'); 
     foreach($fields as $field) {
     dprint_r($field);
     $result[$field[data_type] .':' . $field[id]] = 
     implode(' : ', array($field[name], $field[description]));
     }
     break;
    */
    
    /* disabled for now
     case 'search':
     require_once 'CRM/Core/PseudoConstant.php';
     $result =& CRM_Core_PseudoConstant::savedSearch();
     break; */
    
  }
  //For now, if we had to restrict the query, do not cache
  if (!$params)
    $cache[$type] = $result;

  return $result;
}



/**
 * Forms API Widget for selecting groups
 *
 * @ingroup CiviNode_Form_Widgets
 * @return array of Forms API parameters
 */
function civinode_group_selector($title, $desc = NULL, $default_gid = 0,
                                 $params = NULL){
  $groups = civinode_get_groups($params);
  if (!$default_gid)
     $default_gid = 0; //must be explicit for Drupal Forms API
  //Since none is an option...
  $groups[0] = t('<select a group>');
  $widget = array(
                  '#type' => 'select',
                  '#title' => $title,
                  '#options' => $groups
                  );
  //if ($default_gid)
  $widget['#default_value'] = $default_gid;
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;

}


/**
 * Forms API Widget for selecting CiviCRM profiles
 *
 * @ingroup CiviNode_Form_Widgets
 * @return array of Forms API parameters
 */
function civinode_profile_selector($title, $desc = NULL, $default_pid = 0,
                                   $params = NULL){
  $options = civinode_get_profiles($params);
  $options[0] = t('-- select a profile --  ');
  $widget = array(
                  '#type' => 'select',
                  '#title' => $title,
                  '#options' => $options
                  );
  if ($default_pid)
    $widget['#default_value'] = $default_pid;
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;


}

/**
 * Campaign selector...
 */
function civinode_voter_campaign_selector(){
	if (civinode_voter_install())
	  return crm_voter_get_campaigns();
	else
	  return array();
}


/**
 * Canvass selector...
 */
function civinode_voter_canvass_selector(){
	if (civinode_voter_install())
	  return crm_voter_get_canvasses();
	else
	  return array();
}

/**
 * A forms api widget for a autocompleting contact widget
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_contact_auto_selector($title, $desc = NULL,
                                        $autocomplete = 'civinode/autocomplete/contact',
                                        $key='any', $val = 'any'){
  $widget = array(
                  '#type' => 'textfield',
                  '#title' => $title,
                  '#size' => 60,
                  '#autocomplete_path' => $autocomplete . "/$key/$val"
                  );
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;


}


/**
 * Editing widget for a profile, wrapped for Forms API
 *
 * For now, we directly use CiviCRM UI, since we don't have quite
 * enough metadata available to us to do this right ourselves.
 *
 * If called w/o a contact_id, this widget will put up the UI for
 * creating a new contact.
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_get_edit_profile_html($pid, $contact_id = 0){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }

  $action = $contact_id ? 
    CIVINODE_MAGIC_ACTION_UPDATE : 
    CIVINODE_MAGIC_ACTION_ADD;

  if (!isset($_POST['op']) or $_POST['op'] == t('Delete') 
      or $_POST['op'] == t('Preview')) {
    //$action += CIVINODE_MAGIC_ACTION_PREVIEW;
    unset($_POST['_qf_default']);
    unset($_REQUEST['_qf_default']);
  }
  $title = crm_uf_get_profile_title($pid);
  $reset = isset($_POST['edit']) ? FALSE : TRUE;
  $output = crm_uf_get_profile_html( $contact_id, $title, $action, FALSE, $reset );
  //Now wrap it for the API.  It may make sense
  //to put this into a fieldset as well.
  return array('its_data' => array('#value' => $output),
               '#title' => $title,
               '#collapsible' => TRUE,
               '#type' => 'fieldset',
               '#weight' => -3);

}


/**
 * Handlers for relationships
 * @defgroup CiviCRM_Relationships
 */



/**
 * Fetch a relationship for a contact
 *
 * @ingroup CiviCRM_Relationships
 * @param int $contact_id CRM contact_id
 * @param array $rel_types names of relation types to retrieve.
 * @return
 */
function civinode_fetch_relationships($contact_id, $rel_types = array()){
  static $types;
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), 
             WATCHDOG_ERROR);
    return NULL;
  }

  //No contact, no relationship!
  if (!$contact_id)
    return FALSE;
  $contact = crm_get_contact(array('contact_id' => $contact_id));
  if (!$contact or civinode_check_error($contact))
    return FALSE;
  //First, we need type objects.  Cache 'em
  $type_objs = array();
  if ($rel_types and count($rel_types)) {
    if (!$types) {
      $all_types = crm_get_relationship_types();
      foreach ($all_types as $robj) {
        if (isset($robj->name_a_b)) {
          $types[$robj->name_a_b] = $robj;
          if ($robj->name_a_b != $robj->name_b_a)
            $types[$robj->name_b_a] = $robj;
        }
      }
    }
    foreach ($rel_types as $type_name) {
      //pull out the right objects.  For now, a missing
      //type is *not* an error (TO DO: best behavior?)
      if (isset($types[$type_name]))
        $type_objs[] = $types[$type_name];
    }
  }
  //Docs say this returns an array or null
  return crm_get_relationships($contact, NULL, $type_objs);

}

/**
 * Dojo support
 */

function civinode_contact_autoselect_widget($var_name, $title, 
                                        $key = '', $val = '', 
                                        $group_id = 0,
                                        $desc = NULL, 
                                        $initialize = TRUE) {
  static $dojo_init, $dojo_cache;
  $self_id = "cbox_{$var_name}";
  $shadow_id = "cbox_val_{$var_name}";
  $setter_func = "cn_cck_{$var_name}_setter";
  $attribs =
    array('dojoType' => 'ComboBox',
          'mode' => 'remote',
          'dataUrl' => _civinode_dojo_dataurl($group_id),
          'onValueChanged' => $setter_func,
          //'setSelectedValue' => "$setter_func(arguments[0])",
    );
   $item['cbox']['key'] =
     array('#type' => 'textfield',
           '#title' => $title,
           '#id' => $self_id,
           '#description' => $desc,
           '#maxlength' => 60,
           '#attributes' => $attribs,
           '#default_value' => $key, //sort name of contact
     );
   //shadow item to hold val info
   $attribs = 
     array('style' => 'display: none;');
   $item['cbox']['value'] =
     array(
       '#type' => 'textfield',
       '#attributes' => $attribs,
       '#id' => $shadow_id,
       '#default_value' => $val,
     );
  if ($initialize and !$dojo_init) {
    $dojo_init = TRUE;
    $dojo_cache = array();
    $dojo_lib = _civinode_get_path('dojo_lib');
    $configs = "
    var djConfig = { isDebug: false };
    ";
    drupal_add_js($configs, 'inline', 'header');
    //add the lib file, cache it
    drupal_add_js($dojo_lib, 'module', 'footer');
    //activate the widget
    $js = "
    //alert('test this');
    dojo.require('dojo.widget.ComboBox');
    ";
    drupal_add_js($js, 'inline', 'footer');
  }
  
  //and make sure setters get written exactly once
  if (!isset($dojo_cache[$setter_func])) {
    $dojo_cache[$setter_func] = 1;
    _civinode_dojo_render_callback_js($setter_func, $shadow_id, $self_id, $val, $key);
  }
  return $item;
}

function _civinode_dojo_render_callback_js($funcname, $targ_did, $self_id, $sort_name, $contact_id) {
  //TODO add cslashes
  $js = "
  function $funcname(value) {
    var control_did = '$targ_did';
    var handle = document.getElementById(control_did);
    if (handle) {
      handle.value = value;
    }
  }
  
  function init_$self_id(){
    dojo.widget.byId('$self_id').setAllValues('$sort_name', '$contact_id');
    document.getElementById('$targ_did').value = '$sort_name';
  }
  ";
  //render this into the header
  drupal_add_js($js, 'inline');
  if ($sort_name) {
    $init_js = "
    dojo.addOnLoad(init_$self_id);
    ";
    //and add it to run deferred
    drupal_add_js($init_js, 'inline', 'footer', TRUE);
  }
}
/**
 * Put path info in one place (since this may change)
 *
 * @param string $item
 */
function _civinode_get_path($item = 'base') {
  global $civicrm_root;
  $docroot = $_SERVER['DOCUMENT_ROOT'];
  $len = strlen($docroot);
  $base = substr($civicrm_root, $len + 1);
  switch ($item) {
    case 'ajax':
      return $base . "extern/ajax.php";
    case 'dojo_lib':
      return $base . "packages/dojo/dojo.js";
    default:
      return $base;
  }
}

function _civinode_dojo_dataurl($group_id = 0) {
  //$path = _civinode_get_path('ajax');
  //TODO generalized domain_id as set below
  $path = url("civinode/dojo/contact/$group_id");
  return $path . "?d=1&s=%{searchString}";
}


function civinode_dojo_widgets($widget_type, $group_id = 0, $size = 6) {
  if (_civinode_cck_initialize() and $widget_type == 'contact') {
    require_once _civinode_get_path() . "CRM/Utils/Type.php";
    $domainID = CRM_Utils_Type::escape( $_GET['d'], 'Integer' );
    $name     = strtolower( CRM_Utils_Type::escape( $_GET['s'], 'String'  ) );

    $elements = _civinode_contact_list_by_group($group_id, $size, $name);
    //$cnt = count($elements);
    //error_log("$cnt elements for grp $group_id and $name");
    require_once 'Services/JSON.php';
    $json =& new Services_JSON( );
    echo $json->encode( $elements );
  }
  exit();
}

function civinode_group_contact_popup($add_empty = TRUE, $group_id = 0, $size = 10) {
  _civinode_cck_initialize();
  $elements = _civinode_contact_list_by_group($group_id, $size);
  $options = array();
  if ($add_empty) {
    $options['0'] =  t('--select contact--');
  }
  foreach ($elements as $pair) {
    $options[$pair[1]] = $pair[0];
  }
  return $options;
}


function _civinode_contact_list_by_group($group_id, $size, $name = '') {
  //error_log('invoke dojo search');
  $params = array();
  if ($group_id)
  $params['group'] = array($group_id => 1);

  //Initialize the contact query
  if (!function_exists('crm_create_event')) {
    //we are running < 1.7, likely 1.6
    require_once 'CRM/Contact/Form/Search.php';
    $newP =& CRM_Contact_Form_Search::convertFormValues( $params );
  }
  else {
    //1.7 version
    //require_once 'CRM/Contact/BAO/Query.php';
    $newP =& CRM_Contact_BAO_Query::convertFormValues( $params );
    
  }
  $query =& new CRM_Contact_BAO_Query( $newP );
  list( $select, $from, $where ) = $query->query( );
  //simplify and add our restrictions
  $select = "SELECT contact_a.sort_name, contact_a.id";
  if ($name) {
    if (!$where)
      $where = "WHERE ";
    else
      $where .= " AND ";
    $where .= "contact_a.sort_name LIKE '$name%'";
  }
  $sql = "$select $from $where ORDER BY contact_a.sort_name LIMIT $size";
  //error_log($sql);
  $nullArray = array( );
  $dao = CRM_Core_DAO::executeQuery( $sql, $nullArray );
  $count = 0;
  $elements = array();
  while ( $dao->fetch( ) && $count < $size - 1 ) {
    $elements[] = array( $dao->sort_name, $dao->id );
    $count++;
  }
  return $elements;
}


function civinode_dojo_widgets_crm_clone($widget_type, $group_id = 0) {
  if (_civinode_cck_initialize() and $widget_type == 'contact') {
    require_once _civinode_get_path() . "CRM/Utils/Type.php";
    $domainID = CRM_Utils_Type::escape( $_GET['d'], 'Integer' );
    $name     = strtolower( CRM_Utils_Type::escape( $_GET['s'], 'String'  ) );
    //error_log('invoke dojo search');
    $query = "
SELECT sort_name, id
  FROM civicrm_contact
 WHERE domain_id = $domainID
   AND sort_name LIKE '$name%'
ORDER BY sort_name
LIMIT 6";

    $nullArray = array( );
    $dao = CRM_Core_DAO::executeQuery( $query, $nullArray );

    $count = 0;
    $elements = array();
    while ( $dao->fetch( ) && $count < 5 ) {
        $elements[] = array( $dao->sort_name, $dao->id );
        $count++;
    }

    require_once 'Services/JSON.php';
    $json =& new Services_JSON( );
    echo $json->encode( $elements );

  }
  exit();
}
